from . import utils
from . import soup
from ..task import Task

from sphinx.util.nodes import set_source_info
from sphinx.util.docutils import SphinxDirective
from docutils import nodes



from sphinx.util import logging
_logger = logging.getLogger(__name__)


def setup(app):
    app.add_directive('ot-taskmeta', _TaskMetaDirective)
    app.connect('doctree-resolved', _ev_doctree_resolved__expand_taskmeta_nodes)

def _ev_doctree_resolved__expand_taskmeta_nodes(app, doctree, docname):
    soup.sphinx_create_soup(app)
    for n in doctree.traverse(_TaskMetaNode):
        task = app.ot_soup.element_by_path(n.path)
        assert isinstance(task, Task)

        # simplicity: show task metadata in a bullet list that contain
        # all that's necessary but is not overly beautiful.

        # TODO: find out how to create tables

        bl = nodes.bullet_list()

        # title
        elems = [nodes.Text(f'Title: {task.title}')]
        p = nodes.paragraph()
        p += elems
        li = nodes.list_item()
        li += p
        bl += li

        # path
        elems = [nodes.Text(f'Path: {".".join(task.path)}')]
        p = nodes.paragraph()
        p += elems
        li = nodes.list_item()
        li += p
        bl += li

        # ref
        elems = [nodes.Text('Ref: ')]
        ref = nodes.reference()
        ref['refuri'] = app.builder.get_relative_uri(
            from_=docname, to=task.docname)
        ref += nodes.Text('.'.join(task.path))
        elems.append(ref)

        p = nodes.paragraph()
        p += elems
        li = nodes.list_item()
        li += p
        bl += li

        # implementation_points
        elems = [nodes.Text(f'Implementation Points: {task.implementation_points}')]
        p = nodes.paragraph()
        p += elems
        li = nodes.list_item()
        li += p
        bl += li

        # documentation_points
        elems = [nodes.Text(f'Documentation Points: {task.documentation_points}')]
        p = nodes.paragraph()
        p += elems
        li = nodes.list_item()
        li += p
        bl += li

        # integration_points
        elems = [nodes.Text(f'Integration Points: {task.integration_points}')]
        p = nodes.paragraph()
        p += elems
        li = nodes.list_item()
        li += p
        bl += li

        n.replace_self(bl)
 
class _TaskMetaNode(nodes.Element):
    def __init__(self, path):
        super().__init__(self)
        self.path = path

class _TaskMetaDirective(SphinxDirective):
    required_arguments = 1   # path

    def run(self):
        path = utils.element_path(self.arguments[0].strip())

        l = _TaskMetaNode(path=path)
        l.document = self.state.document
        set_source_info(self, l)

        return [l]




class OpenTrainingError(Exception):
    '''Base class for all OpenTraining errors'''

class BadPath(OpenTrainingError):
    pass

class ElementError(OpenTrainingError):
    '''Base class for errors tied to a single element'''
    def __init__(self, msg, element):
        super().__init__(msg)
        self.element = element

class DependencyError(ElementError):
    pass

class PathNotFound(ElementError):
    pass

class NotCommitted(ElementError):
    pass

class AlreadyCommitted(ElementError):
    pass

class CompoundError(OpenTrainingError):
    '''An error that *contains* multiple errors. Used in situations where
    we do not bail out early, but rather continue, and represent the
    user with a list of collected error messages.  '''

    def __init__(self, msg, errors):
        super().__init__(msg)
        self.errors = errors



from .node import Node


class Task(Node):
    def __init__(self, 
                 title, path, docname, 
                 dependencies, userdata,

                 implementation_points, implementors,
                 documentation_points, documenters,
                 integration_points, integrators):

        if len(implementors) != 0:
            assert sum(share for person, share in implementors) == 100
        if len(documenters) != 0:
            assert sum(share for person, share in documenters) == 100
        if len(integrators) != 0:
            assert sum(share for person, share in integrators) == 100

        super().__init__(
            title=title, 
            path=path, 
            docname=docname, 
            # add persons as task dependencies
            dependencies=dependencies + [person for person, share in implementors + documenters + integrators], 
            userdata=userdata)

        self.implementation_points = implementation_points
        self.documentation_points = documentation_points
        self.integration_points = integration_points

        self.implementors = implementors
        self.documenters = documenters
        self.integrators = integrators

    def __str__(self):
        return 'Task:'+super().__str__()

    def resolve_paths(self, soup):
        resolved_implementors = []
        for person_path, share in self.implementors:
            person = soup.element_by_path(person_path)
            resolved_implementors.append((person, share))
        resolved_documenters = []
        for person_path, share in self.documenters:
            person = soup.element_by_path(person_path)
            resolved_documenters.append((person, share))
        resolved_integrators = []
        for person_path, share in self.integrators:
            person = soup.element_by_path(person_path)
            resolved_integrators.append((person, share))

        self.implementors = resolved_implementors
        self.documenters = resolved_documenters
        self.integrators = resolved_integrators



from .node import Node


class Person(Node):  # is-a Node only because that holds path,
                     # docname, and userdata
    def __init__(self, title, path, docname, userdata, firstname,
                 lastname):
        super().__init__(
            title=title, 
            path=path, 
            docname=docname, 
            userdata=userdata,
            dependencies=[], 
            )
        self.firstname = firstname
        self.lastname = lastname

    def __str__(self):
        return f'Person:{str(super())}, firstname:{self.firstname}, lastname:{self.lastname}'



from ..soup import Soup
from ..element import Element
from ..topic import Topic
from ..exercise import Exercise
from ..task import Task
from ..group import Group
from .. import errors

from sphinx.util import logging
_logger = logging.getLogger(__name__)


def _prepare_app(app):
    if hasattr(app, 'ot_soup'):
        raise OpenTrainingError('Soup already created, cannot add one more element')
    if not hasattr(app.env, 'ot_elements'):
        app.env.ot_elements = {}

def sphinx_add_element(app, element):
    _prepare_app(app)
    assert isinstance(element, Element)
    app.env.ot_elements[element.docname] = element

def sphinx_purge_doc(app, env, docname):
    if hasattr(env, 'ot_elements'):
        env.ot_elements.pop(docname, None)

def sphinx_create_soup(app):
    if hasattr(app, 'ot_soup'):
        return

    app.ot_soup = Soup()
    for element in app.env.ot_elements.values():
        app.ot_soup.add_element(element)

    try:
        app.ot_soup.commit()
    except errors.CompoundError as e:
        for err in e.errors:
            if isinstance(err, errors.ElementError):
                _logger.warning(str(err), location=err.element.userdata)
            else:
                assert False, f'hm. cannot handle {err}'
            



from . import utils
from . import soup
from .. import errors
from ..topic import Topic
from ..group import Group
from ..errors import OpenTrainingError

from sphinx.util.docutils import SphinxDirective
from sphinx.util.nodes import set_source_info
from sphinx.util import logging
from docutils import nodes

from networkx.algorithms.dag import topological_sort

logger = logging.getLogger(__name__)


def setup(app):
    app.add_directive('ot-group', _GroupDirective)
    app.connect('doctree-read', _ev_doctree_read__extract_groupnodes)

def _ev_doctree_read__extract_groupnodes(app, doctree):
    '''Add group metadata to soup. Leave group node intact; it is expanded
    later when all metadata has been collected.'''

    try:
        docname = app.env.docname
        group_nodes = list(doctree.traverse(_GroupNode))
        if len(group_nodes) > 1:
            raise OpenTrainingError(f'{docname} contains multiple groups')

        for n in group_nodes:
            soup.sphinx_add_element(app, Group(
                docname=docname, 
                title=utils.get_document_title(docname, doctree), 
                path=n.path,
                userdata=n,
            ))
            n.replace_self([])
    except Exception:
        logger.exception(f'{docname}: cannot extract group nodes')
        raise

class _GroupNode(nodes.Element):
    def __init__(self, path):
        super().__init__(self)
        self.path = path

class _GroupDirective(SphinxDirective):
    required_arguments = 1   # path

    def run(self):
        path = utils.element_path(self.arguments[0].strip())

        group = _GroupNode(path=path)
        group.document = self.state.document
        set_source_info(self, group)

        return [group]




from . import utils
from . import soup
from .. import errors
from ..topic import Topic
from ..exercise import Exercise
from ..task import Task
from ..person import Person
from ..node import Node
from ..group import Group

from sphinx.util.docutils import SphinxDirective
from sphinx.util.nodes import set_source_info
from sphinx.util import logging
from docutils import nodes

import subprocess
import re

_logger = logging.getLogger(__name__)


def setup(app):
    app.add_directive('ot-graph', _GraphDirective)
    app.connect('doctree-resolved', _ev_doctree_resolved__expand_topicgraph_nodes)

def _ev_doctree_resolved__expand_topicgraph_nodes(app, doctree, docname):
    '''"doctree-resolved" event handler to expand topic graph nodes'''
    try:
        soup.sphinx_create_soup(app)
        expander = _GraphExpander(app=app, docname=docname)
        for n in doctree.traverse(_GraphNode):
            expander.expand(n)
    except Exception as err:
        _logger.warning(f'{docname}: cannot expand topic graph', location=err.element.userdata)

class _GraphNode(nodes.Element):
    def __init__(self, entries):
        super().__init__(self)
        self.entries = entries

class _GraphDirective(SphinxDirective):
    option_spec = {
        'entries': utils.list_of_elementpath,
    }
    def run(self):
        node = _GraphNode(entries=self.options.get('entries', []))
        node.document = self.state.document
        set_source_info(self, node)
        return [node]

class _GraphExpander:
    def __init__(self, app, docname):
        self._app = app
        self._docname = docname

    def expand(self, node):
        graph, hilit_nodes = self._graphnode_to_graph(node)
        dot = self._graph_to_dot(graph=graph, hilit_nodes=hilit_nodes)
        svg = self._dot_to_svg(dot=dot)
        node.replace_self(nodes.raw(svg, svg, format='html'))

    def _graphnode_to_graph(self, node):
        assert isinstance(node, _GraphNode)

        worldgraph = self._app.ot_soup.worldgraph()
        if len(node.entries) == 0:
            return worldgraph, set()

        # highlight node-type entries. expand group-type entries as
        # their contained nodes.
        node_entries = set()
        hilit_nodes = set()

        for entry_path in node.entries:
            entry = self._app.ot_soup.element_by_path(entry_path)
            if isinstance(entry, Node):
                node_entries.add(entry)
                hilit_nodes.add(entry)
            elif isinstance(entry, Group):
                for _, elem in entry.iter_recursive():
                    if isinstance(elem, Node):
                        node_entries.add(elem)
            else:
                assert False, entry_path

        return self._app.ot_soup.subgraph(node_entries), hilit_nodes

    def _graph_to_dot(self, graph, hilit_nodes):
        lines = [
            'digraph {',
        ]

        root_cluster = self._dot_group_clusters(graph)
        lines.extend(self._dot_cluster_lines(root_cluster, hilit_nodes=hilit_nodes))

        for src, dst in graph.edges:
            lines.extend(self._dot_edge_lines(src, dst))

        lines.append('}')

        return '\n'.join(lines)
    
    class Cluster:
        def __init__(self, group):
            self.group = group
            self.clusters = []
            self.nodes = []  # leaf topics

    def _dot_group_clusters(self, graph):
        root_cluster = self.Cluster(self._app.ot_soup.root)

        have_clusters = { self._app.ot_soup.root: root_cluster }
        # walk topics and create cluster hierarchy from their
        # containing groups
        for n in graph.nodes:
            assert isinstance(n, Node), n
            cluster = self._dot_make_cluster(n.parent, have_clusters)
            cluster.nodes.append(n)

        return root_cluster

    def _dot_make_cluster(self, group, have_clusters):
        assert type(group) is Group

        cluster = have_clusters.get(group)
        if not cluster:
            cluster = self.Cluster(group)
            have_clusters[group] = cluster
            parent_cluster = self._dot_make_cluster(group.parent, have_clusters)
            parent_cluster.clusters.append(cluster)

        return cluster

    def _dot_cluster_lines(self, cluster, hilit_nodes):
        lines = []
        if cluster.group is not self._app.ot_soup.root:
            lines.append('subgraph cluster_' + self._dot_id_from_path(cluster.group.path) + '{')
            lines.append(f'label = "{cluster.group.title}";')
            lines.append('style = rounded;')  # rounded corners

        for n in cluster.nodes:
            lines.extend(self._dot_node_lines(n, hilit=(n in hilit_nodes)))
        for subcluster in cluster.clusters:
            lines.extend(self._dot_cluster_lines(subcluster, hilit_nodes))

        if cluster.group is not self._app.ot_soup.root:
            lines.append('}')
        return lines

    @staticmethod
    def _percent_to_rgb(percent):
        # colors taken from
        # https://www.w3schools.com/colors/colors_picker.asp,
        # "Lightness"
        if percent == 0:
            return "#ffffff"
        elif 0 < percent <= 10:
            return "#e6ffe6"
        elif 10 < percent <= 20:
            return "#ccffcc"
        elif 20 < percent <= 30:
            return "#b3ffb3"
        elif 30 < percent <= 40:
            return "#99ff99"
        elif 40 < percent <= 50:
            return "#80ff80"
        elif 50 < percent <= 60:
            return "#66ff66"
        elif 60 < percent <= 70:
            return "#4dff4d"
        elif 70 < percent <= 80:
            return "#33ff33"
        elif 80 < percent <= 90:
            return "#1aff1a"
        elif 90 < percent <= 100:
            return "#00ff00"

        assert False, f'invalid percentage: {percent}'

    def _dot_node_lines(self, node, hilit):
        uri = self._app.builder.get_relative_uri(from_=self._docname, to=node.docname)
        node_id = '_'.join(node.path)

        border = 1
        if hilit:
            border *= 3

        if isinstance(node, Topic):
            return [
                f'{node_id} [',
                f'    label="{node.title}";',
                f'    href="{uri}";',
                '    style=filled;',
                f'    penwidth="{border}";',
                '    fillcolor="#DCDCDC";'
                '];',
            ]
        elif isinstance(node, Exercise):
            return [
                f'{node_id} [',
                f'    label="{node.title}";',
                f'    href="{uri}";',
                '    style=filled;',
                f'    penwidth="{border}";',
                '    fillcolor="red";'
                '];',
            ]
        elif isinstance(node, Task):
            label = '{'
            label += node.title
            label += '}'

            return [
                f'{node_id} [',
                f'    label="{label}";',
                f'    href="{uri}";',
                '    shape=Mrecord;',
                '    style=filled;',
                f'    penwidth="{border}"',
#                f'    fillcolor="{self._percent_to_rgb(node.percent_done)}";'
                '];',
            ]

        elif isinstance(node, Person):
            label = '{'
            label += node.title
            label += '|'
            label += f'Firstname: {node.firstname}'
            label += '|'
            label += f'Lastname: {node.lastname}'

            label += '}'

            return [
                f'{node_id} [',
                f'    label="{label}";',
                f'    href="{uri}";',
                '    shape=Mrecord;',
                '    style=filled;',
                f'    penwidth="{border}";'
                '];',
            ]

    def _dot_edge_lines(self, src, dst):
        src_id = '_'.join(src.path)
        dst_id = '_'.join(dst.path)
        return [f'{src_id} -> {dst_id};']

    _re_width = re.compile(r'width\s*=\s*".*"')
    def _dot_to_svg(self, dot):
        try:
            completed = subprocess.run(
                ['dot', '-v', '-T', 'svg'],
                input=dot, check=True, text=True, capture_output=True)
        except subprocess.CalledProcessError as e:
            raise errors.OpenTrainingError(f'dot exited with status {e.returncode}:\n[dot]\n{dot}\n[stderr]\n{e.stderr}')
    
        svg = completed.stdout
        # strip XML declaration (we are embedding it)
        svg = svg[svg.index('<svg'):]

        if True:
            # patch "width" out (we want it scaled to fit the page)
            width = self._re_width.search(svg)
             # hm. if not there, dot has changed, apparently. fix that.
            assert width is not None
            svg = svg[:width.start()] + svg[width.end():]

        return svg

    def _dot_id_from_path(self,  path):
        return '_'.join(path)



from .element import Element


class Node(Element):
    def __init__(self, title, path, docname, dependencies, userdata):
        super().__init__(
            title=title, 
            path=path, 
            docname=docname, 
            userdata=userdata)
        self.dependencies = dependencies



from . import utils
from . import soup
from .. import errors
from ..topic import Topic
from ..group import Group
from ..errors import OpenTrainingError

from sphinx.util.docutils import SphinxDirective
from sphinx.util.nodes import set_source_info
from sphinx.util import logging
from docutils import nodes

from networkx.algorithms.dag import topological_sort

logger = logging.getLogger(__name__)


def setup(app):
    app.add_directive('ot-grouplist', _GroupListDirective)
    app.connect('doctree-resolved', _ev_doctree_resolved__expand_grouplist_nodes)

def _ev_doctree_resolved__expand_grouplist_nodes(app, doctree, docname):
    try:
        soup.sphinx_create_soup(app)
        expander = _GroupListExpander(app=app, docname=docname)
        for n in doctree.traverse(_GroupListNode):
            expander.expand(n)
    except Exception:
        logger.exception(f'{docname}: cannot expand grouplist')
        raise

class _GroupListNode(nodes.Element):
    def __init__(self, path):
        super().__init__(self)
        self.path = path

class _GroupListDirective(SphinxDirective):
    required_arguments = 1   # path

    def run(self):
        path = utils.element_path(self.arguments[0].strip())

        l = _GroupListNode(path=path)
        l.document = self.state.document
        set_source_info(self, l)

        return [l]

class _GroupListExpander:
    def __init__(self, app, docname):
        self._app = app
        self._docname = docname

    def expand(self, node):
        group = self._app.ot_soup.element_by_path(node.path)
        topics = (t for _,t in group.iter_recursive() if isinstance(t, Topic))
        graph = self._app.ot_soup.worldgraph().subgraph(topics)
        topo = topological_sort(graph)

        bl = nodes.bullet_list()
        for topic in reversed(list(topo)):
            if not isinstance(topic, Topic):
                continue
            li = nodes.list_item()
            li += self._topic_paragraph(topic.path)
            bl += li
        node.replace_self(bl)

    def _topic_paragraph(self, path):
        topic = self._app.ot_soup.element_by_path(path)
        assert isinstance(topic, Topic), f'dependency on non-topic {path}?'
        p = nodes.paragraph()
        p += self._topic_headline_elems(path)
        return p

    def _topic_headline_elems(self, path):
        topic = self._app.ot_soup.element_by_path(path)
        elems = []
        elems.append(nodes.Text(f'{topic.title} ('))

        ref = nodes.reference()
        ref['refuri'] = self._app.builder.get_relative_uri(
            from_=self._docname, to=topic.docname)
        ref += nodes.Text('.'.join(topic.path))
        elems.append(ref)
        elems.append(nodes.Text(')'))
        
        return elems



from collections import defaultdict


class Grading:
    def __init__(self, persons, tasks):
        self.persons = persons
        self.tasks = tasks

    def points_per_person(self):
        'Return iterable of (person, points)'

        points_per_person = defaultdict(int)

        for task in self.tasks:
            for person, share in task.implementors:
                points = share * task.implementation_points
                points_per_person[person] += points
            for person, share in task.documenters:
                points = share * task.documentation_points
                points_per_person[person] += points
            for person, share in task.integrators:
                points = share * task.integration_points
                points_per_person[person] += points

        return ((person, points_per_person[person]) for person in self.persons)



from . import errors


class Element:
    '''
    Base of all OpenTraining artifacts (Topic, Exercise, Task, Person, ...)
    '''
    def __init__(self, title, path, docname, userdata):
        ''':param title: string that is sometimes (?) displayed somewhere (?)

        :param docname: leaks out from sphinx/docutils. Sometimes (?)
        used to display error messages. Should be removed in favor of
        ``userdata``

        :param userdata: during sphinx/docutils processing, we try to
        create OpenTraining elements as early as possible while
        keeping reference to the original docutils elements. For
        debugging and error messages only.

        :param path: the requested path across the hierarchy that this
        element should be placed under. Note that the hierarchy is not
        yet in place when an Element is create - this is just a
        request for a later step: when the soup is committed, Group
        elements are created, and is that point where the ``path``
        argument is used to add this Element to a *parent* group.

        '''

        _verify_is_path(path)

        self.title = title
        self.docname = docname
        self.userdata = userdata
        if path:
            self._requested_path = path
        else:   # root group; no parent
            self.parent = None

    def __str__(self):
        if hasattr(self, '_requested_path'):
            return f'{self._requested_path} (uncommitted)'
        else:
            return f'{self.path}'

    @property
    def path(self):
        if hasattr(self, '_requested_path'):
            raise errors.OpenTrainingError(f'{self} is not yet committed (path not known)')
        if self.parent:
            return self.parent.path + [self.parent.element_name(self)]
        else:
            return []

    def resolve_paths(self, soup):
        pass


def _verify_is_path(path):
    if type(path) not in (list, tuple):
        raise errors.BadPath(f'Not a valid path: {path} is neither list nor tuple')
        
    for elem in path:
        if type(elem) is not str:
            raise errors.BadPath(f'Not a valid path: {path} ({elem} is not str)')
        if not elem.isidentifier():
            raise errors.BadPath(f'Not a valid path: {path} ({elem} is not an identifier)')
    



from . import utils
from . import soup
from .. import errors
from ..topic import Topic

from sphinx.util.docutils import SphinxDirective
from sphinx.util.nodes import set_source_info
from sphinx.util import logging
from docutils import nodes

logger = logging.getLogger(__name__)


def setup(app):
    app.add_directive('ot-topic', _TopicDirective)
    app.connect('doctree-read', _ev_doctree_read__extract_topicnodes)

def _ev_doctree_read__extract_topicnodes(app, doctree):
    try:
        docname = app.env.docname
        topic_nodes = list(doctree.traverse(_TopicNode))
        if len(topic_nodes) > 1:
            raise errors.OpenTrainingError(f'{docname} contains multiple topics')

        for n in topic_nodes:
            soup.sphinx_add_element(app, Topic(
                title=utils.get_document_title(docname, doctree),
                path=n.path, 
                docname=docname,
                userdata=n,
                dependencies=n.dependencies,
            ))
            n.replace_self([])
    except Exception:
        logger.exception(f'{docname}: cannot extract topic nodes')
        raise
        
class _TopicNode(nodes.Element):
    def __init__(self, path, dependencies):
        super().__init__(self)
        self.title = None
        self.path = path
        self.dependencies = dependencies

class _TopicDirective(SphinxDirective):
    required_arguments = 1   # path

    option_spec = {
        'dependencies': utils.list_of_elementpath,
    }

    def run(self):
        path = utils.element_path(self.arguments[0].strip())
        dependencies = self.options.get('dependencies', [])

        topic = _TopicNode(path=path, dependencies=dependencies)
        topic.document = self.state.document
        set_source_info(self, topic)

        return [topic]




from .element import Element
from .group import Group
from .node import Node
from .topic import Topic
from .exercise import Exercise
from .task import Task
from . import errors
from . import element

from networkx.algorithms.dag import descendants
from networkx import DiGraph


class Soup:
    def __init__(self):
        self._elements = set()
        self._root_group = None
        self._worldgraph = None

    def __len__(self):
        # well this is a bit dumb
        return len(list(self._root_group.iter_recursive()))

    def add_element(self, element):
        assert isinstance(element, Element)
        self._assert_uncommitted()
        if self._worldgraph:
            raise errors.OpenTrainingError(f'cannot add {element}: graph already made')
        self._elements.add(element)

    def commit(self):
        if self._root_group is not None:
            return

        elements_to_resolve = list(self._elements)

        # build up hierarchy (thereby emptying self._elements)
        self._root_group = Group(
            title='Root', 
            path=(), 
            docname='', 
            userdata=None,   # no associated docutils node
        )
        self._make_hierarchy()
        self._add_nodes_to_groups()
        assert len(self._elements) == 0, self._elements

        # once the elements have paths in their final hierarchy, we
        # can let them resolve their own stuff. for example, a task
        # initially refers to a person's *path* - final situation
        # should be though that a task refers to the person directly.
        for element in elements_to_resolve:
            element.resolve_paths(self)

        del self._elements
        self.worldgraph()    # only to detect missing dependencies
                             # early

    @property
    def root(self):
        return self._root_group

    def element_by_path(self, path):
        return self._root_group.element_by_path(path)

    def worldgraph(self):
        self._assert_committed()
        return self._make_worldgraph()

    def subgraph(self, entrypoints):
        '''Given entrypoints, compute a subgraph of the world graph that
        contains the entrypoints and all their descendants.

        entrypoints is an iterable of element paths or elements (can
        be mixed)
        '''

        # paranoia
        for e in entrypoints:
            assert isinstance(e, Element)

        self._assert_committed()
        world = self._make_worldgraph()

        topics = set()
        for topic in entrypoints:
            topics.add(topic)
            topics.update(descendants(world, topic))
        return world.subgraph(topics)

    def _make_worldgraph(self):
        if self._worldgraph is not None:
            return self._worldgraph

        collected_errors = []
        self._worldgraph = DiGraph()
        for name, elem in self._root_group.iter_recursive():
            if not isinstance(elem, Node):
                continue
            self._worldgraph.add_node(elem)
            for target_path in elem.dependencies:
                try:
                    target_topic = self.element_by_path(target_path)
                    self._worldgraph.add_edge(elem, target_topic)
                except errors.PathNotFound as e:
                    collected_errors.append(
                        errors.DependencyError(
                            f'{elem.docname} ({elem}): dependency {target_path} not found', 
                            element=elem))

        if len(collected_errors) != 0:
            raise errors.CompoundError('cannot build world graph', errors=collected_errors)
        return self._worldgraph

    def _make_hierarchy(self):
        level = 1
        while True:
            all_groups = [g for g in self._elements if isinstance(g, Group)]
            if not all_groups:   # no more groups
                break
            level_groups = [g for g in all_groups if len(g._requested_path) == level]
            for g in level_groups:
                self._root_group.add_element(g)
                self._elements.remove(g)
            level += 1

    def _add_nodes_to_groups(self):
        nodes = [n for n in self._elements if isinstance(n, Node)]
        for n in nodes:
            self._root_group.add_element(n)
            self._elements.remove(n)
            
    def _assert_committed(self):
        if self._root_group is None:
            raise errors.NotCommitted('soup not committed')

    def _assert_uncommitted(self):
        if self._root_group is not None:
            raise errors.AlreadyCommitted('soup already committed')



from .node import Node


class Exercise(Node):
    def __init__(self, title, path, docname, dependencies, userdata):
        super().__init__(
            title=title, 
            path=path, 
            docname=docname, 
            dependencies=dependencies, 
            userdata=userdata)

    def __str__(self):
        return 'Exercise:'+super().__str__()



from .sphinxglue import topic
from .sphinxglue import exercise
from .sphinxglue import task
from .sphinxglue import taskmeta
from .sphinxglue import person
from .sphinxglue import grading
from .sphinxglue import group
from .sphinxglue import grouplist
from .sphinxglue import graph
from .sphinxglue import soup
from .sphinxglue import dia

def setup(app):
    app.connect('env-purge-doc', soup.sphinx_purge_doc)

    topic.setup(app)
    exercise.setup(app)
    task.setup(app)
    taskmeta.setup(app)
    person.setup(app)
    grading.setup(app)
    group.setup(app)
    grouplist.setup(app)
    graph.setup(app)
    dia.setup(app)





from sphinx.errors import ExtensionError
from sphinx.transforms.post_transforms.images import ImageConverter
from sphinx.util import logging, images

import subprocess
from subprocess import CalledProcessError, PIPE


logger = logging.getLogger(__name__)

class DiaConverter(ImageConverter):
    conversion_rules = [
        ('image/x-dia', 'image/svg+xml'),
    ]

    def is_available(self):
        cmd = ['dia', '--version']
        try:
            logger.debug(f'Invoking {cmd} ...')
            subprocess.run(cmd, stdout=PIPE, stderr=PIPE, check=True)
            return True
        except OSError:
            logger.warning(f'convert command {cmd} cannot be run.')
            return False
        except CalledProcessError as exc:
            logger.warning(f'{cmd} exited with error:\n'
                           '[stderr]\n%r\n[stdout]\n%r',
                           exc.stderr, exc.stdout)
            return False

    def convert(self, _from, _to):
        cmd = ['dia', '--export', _to, '--filter', 'svg', _from]
        try:
            logger.debug(f'Invoking {cmd} ...')
            subprocess.run(cmd, stdout=PIPE, stderr=PIPE, check=True)
            return True
        except OSError:
            logger.warning('{cmd} failed')
            return False
        except CalledProcessError as exc:
            raise ExtensionError(f'{cmd} exited with error:\n'
                                 '[stderr]\n%r\n[stdout]\n%r' %
                                 (exc.stderr, exc.stdout))

def setup(app):
    # I found this the
    # simplest. (sphinx.environment.collectors.asset.ImageCollector
    # appears to add someting to node['candidates'], but I don't get
    # it.)
    assert '.dia' not in images.mime_suffixes
    images.mime_suffixes['.dia'] = 'image/x-dia'

    app.connect('builder-inited', _debug)
    app.add_post_transform(DiaConverter)

    return {
        'version': 'builtin',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }

def _debug(app):
    # print('DEBUG', app.builder.supported_image_types)
    pass



from . import errors
from .element import Element
from .node import Node


class Group(Element):
    def __init__(self, title, path, docname, userdata):
        super().__init__(
            title=title, 
            path=path, 
            docname=docname, 
            userdata=userdata)
        self._children = {}    # {name: element}

    def __str__(self):
        return 'Group:'+super().__str__()

    def __len__(self):
        return len(self._children)

    def add_element(self, element):
        child_name = element._requested_path[0]
        if len(element._requested_path) == 1: # leaf; add to children
            child = self._children.get(child_name)
            if child:
                raise errors.OpenTrainingError(f'{self}: cannot add "{child_name}"; already exists: {child}')
            self._children[child_name] = element
            del element._requested_path
            element.parent = self
        else:
            parent = self._children.get(child_name)
            if parent is None:
                raise errors.OpenTrainingError(f'{self}: cannot add "{element._requested_path}": '
                                        f'intermediate "{child_name}" does not exist')
            element._requested_path = element._requested_path[1:]
            parent.add_element(element)

    def element_by_path(self, path):
        '''Get element by path. path is relative to this group.'''

        element = self._children.get(path[0])
        if element is None:
            raise errors.PathNotFound(f'{self}: no element with name "{path[0]}"', element=element)
        if len(path) == 1:
            return element

        return element.element_by_path(path[1:])

    def child_by_name(self, name):
        '''Get direct child element by name.'''
        child = self._children.get(name)
        if child is None:
            raise errors.OpenTrainingError(f'{self}: no child with name {name}')
        return child

    def element_name(self, element):
        '''Get name of element under which it is know to his group.

        Raises OpenTrainingError if element not there.

        '''
        for name, elem in self._children.items():
            if elem is element:
                return name
        else:
            raise errors.OpenTrainingError(f'{element} is not a child of {self}')

    def has_element(self, element):
        '''Does this group have the element?'''
        for name, elem in self._children.items():
            if elem is element:
                return True
        return False

    def iter_children(self):
        '''Iterator (name, element) over direct children'''
        yield from self._children.items()

    def iter_recursive(self):
        '''Iterator (name, element) over descendants, recursively'''
        for name, elem in self._children.items():
            if isinstance(elem, Node):
                yield name, elem
            elif isinstance(elem, Group):
                yield name, elem
                #yield from elem.iter_children_dfs()
                for n,e in elem.iter_recursive():
                    yield n,e
            else:
                assert False, f'invalid element type {type(e)}'



from . import utils
from . import soup
from .. import errors
from ..person import Person

from sphinx.util.docutils import SphinxDirective
from sphinx.util.nodes import set_source_info
from sphinx.util import logging
from docutils import nodes

logger = logging.getLogger(__name__)


def setup(app):
    app.add_directive('ot-person', _PersonDirective)
    app.connect('doctree-read', _ev_doctree_read__extract_personnodes)

def _ev_doctree_read__extract_personnodes(app, doctree):
    try:
        docname = app.env.docname
        person_nodes = list(doctree.traverse(_PersonNode))
        if len(person_nodes) > 1:
            raise errors.OpenTrainingError(f'{docname} contains multiple persons')

        for n in person_nodes:
            soup.sphinx_add_element(app, Person(
                docname=docname,
                title=utils.get_document_title(docname, doctree),
                path=n.path, 
                userdata=n,
                firstname=n.firstname,
                lastname=n.lastname,
            ))
            n.replace_self([])
    except Exception:
        logger.exception(f'{docname}: cannot extract person nodes')
        raise
        
class _PersonNode(nodes.Element):
    def __init__(self, path, firstname, lastname):
        super().__init__(self)
        self.title = None
        self.path = path
        self.firstname = firstname
        self.lastname = lastname

class _PersonDirective(SphinxDirective):
    required_arguments = 1   # path

    option_spec = {
        'firstname': str,
        'lastname': str,
    }

    def run(self):
        path = utils.element_path(self.arguments[0].strip())
        firstname = self.options.get('firstname')
        lastname = self.options.get('lastname')

        person = _PersonNode(path=path, 
                             firstname=firstname,
                             lastname=lastname)
        person.document = self.state.document
        set_source_info(self, person)

        return [person]



from . import utils
from . import soup
from .. import errors
from ..task import Task

from sphinx.util.docutils import SphinxDirective
from sphinx.util.nodes import set_source_info
from sphinx.util import logging
from docutils import nodes

logger = logging.getLogger(__name__)


def setup(app):
    app.add_directive('ot-task', _TaskDirective)
    app.connect('doctree-read', _ev_doctree_read__extract_tasknodes)

def _ev_doctree_read__extract_tasknodes(app, doctree):
    try:
        docname = app.env.docname
        task_nodes = list(doctree.traverse(_TaskNode))
        if len(task_nodes) > 1:
            raise errors.OpenTrainingError(f'{docname} contains multiple tasks')

        for n in task_nodes:
            soup.sphinx_add_element(app, Task(
                title = utils.get_document_title(docname, doctree),
                docname = docname,
                path = n.path, 
                dependencies = n.dependencies,
                userdata = n,

                implementation_points = n.implementation_points,
                documentation_points = n.documentation_points,
                integration_points = n.integration_points,
                implementors = n.implementors,
                documenters = n.documenters,
                integrators = n.integrators,
            ))
            n.replace_self([])
    except Exception:
        logger.exception(f'{docname}: cannot extract task nodes')
        raise
        
class _TaskNode(nodes.Element):
    def __init__(self, path, dependencies,
                 implementation_points, documentation_points, integration_points,
                 implementors, documenters, integrators,
                 responsible, initial_estimate, spent, percent_done):
        super().__init__(self)
        self.title = None
        self.path = path

        self.implementation_points = implementation_points
        self.documentation_points = documentation_points
        self.integration_points = integration_points
        self.implementors = implementors
        self.documenters = documenters
        self.integrators = integrators

        self.dependencies = dependencies
        self.responsible = responsible
        self.initial_estimate = initial_estimate
        self.spent = spent
        self.percent_done = percent_done

class _TaskDirective(SphinxDirective):
    required_arguments = 1   # path

    option_spec = {
        'implementation-points': int,
        'documentation-points': int,
        'integration-points': int,
        'implementors': utils.list_of_person_and_share,
        'documenters': utils.list_of_person_and_share,
        'integrators': utils.list_of_person_and_share,
        'dependencies': utils.list_of_elementpath,
        'responsible': str,
        'initial-estimate': int,
        'spent': int,
        'percent-done': int,
    }

    def run(self):
        path = utils.element_path(self.arguments[0].strip())
        dependencies = self.options.get('dependencies', [])

        implementation_points = self.options.get('implementation-points')
        documentation_points = self.options.get('documentation-points')
        integration_points = self.options.get('integration-points')
        implementors = self.options.get('implementors', [])
        documenters = self.options.get('documenters', [])
        integrators = self.options.get('integrators', [])

        responsible = self.options.get('responsible', '')
        initial_estimate = self.options.get('initial-estimate', 0)
        spent = self.options.get('spent', 0)
        percent_done = self.options.get('percent-done', 0)

        task = _TaskNode(path = path, 

                         implementation_points = implementation_points,
                         documentation_points = documentation_points,
                         integration_points = integration_points,
                         implementors = implementors,
                         integrators = integrators,
                         documenters = documenters,

                         dependencies = dependencies,
                         responsible = responsible,
                         initial_estimate = initial_estimate, 
                         spent = spent, 
                         percent_done = percent_done)
        task.document = self.state.document
        set_source_info(self, task)

        return [task]



from . import utils
from . import soup
from ..grading import Grading
from ..errors import OpenTrainingError

from sphinx.util.docutils import SphinxDirective
from sphinx.util.nodes import set_source_info
from docutils import nodes


from sphinx.util import logging
_logger = logging.getLogger(__name__)


def setup(app):
    app.add_directive('ot-pointscollected', _PointsCollectedDirective)
    app.connect('doctree-resolved', _ev_doctree_resolved__expand_pointscollected_nodes)

def _ev_doctree_resolved__expand_pointscollected_nodes(app, doctree, docname):
    for n in doctree.traverse(_PointsCollectedNode):
        persons = []
        tasks = []
        for person in n.persons:
            try:
                persons.append(app.ot_soup.element_by_path(person))
            except OpenTrainingError as e:
                _logger.warning(e, location=n)
        for task in n.tasks:
            try:
                tasks.append(app.ot_soup.element_by_path(task))
            except OpenTrainingError as e:
                _logger.warning(e, location=n)

        grading = Grading(persons = persons, tasks = tasks)
        bl = nodes.bullet_list()
        for person, points in grading.points_per_person():
            elems = [nodes.Text(f'{person.firstname} {person.lastname}: {points}')]
            p = nodes.paragraph()
            p += elems
            li = nodes.list_item()
            li += p
            bl += li
            
        n.replace_self([bl])

class _PointsCollectedNode(nodes.Element):
    def __init__(self, persons, tasks):
        super().__init__(self)
        self.title = None
        self.persons = persons
        self.tasks = tasks

class _PointsCollectedDirective(SphinxDirective):
    required_arguments = 0

    option_spec = {
        'persons': utils.list_of_elementpath,
        'tasks': utils.list_of_elementpath,
    }

    def run(self):
        persons = self.options.get('persons')
        tasks = self.options.get('tasks')

        pointcollected = _PointsCollectedNode(
            persons = persons,
            tasks = tasks,
        )

        pointcollected.document = self.state.document
        set_source_info(self, pointcollected)

        return [pointcollected]



from docutils import nodes


def element_path(pathstr):
    return [c.strip() for c in pathstr.split('.')]

def list_of_elementpath(pathliststr):
    paths = [p.strip() for p in pathliststr.split(',')]
    return [element_path(p) for p in  paths]

def list_of_person_and_share(persons_and_their_shares):
    '''Breaks "group.person_a:70, group.person_b:10" into [(["group", "person_a"], 70), (["group", "person_b"], 70)]
    '''
    ret = []
    for person_and_share in persons_and_their_shares.split(','):
        person_path, sharestr = person_and_share.split(':')
        ret.append((element_path(person_path), int(sharestr)))
    return ret

def get_document_title(docname, doctree):
    for section in doctree.traverse(nodes.section):
        break
    else:
        raise Exception(f'{docname}: no <section> found')
    for title in section.traverse(nodes.title):
        return title.astext()
    else:
        raise Exception(f'{docname}: first <section> has no <title>')



from setuptools import setup

setup(
    name='opentraining',
    version='0.4.1',
    description='A set of Sphinx extensions to manage training material',

    # long_description='blah',
    # long_description_content_type='text/markdown',

    url='https://faschingbauer.me',
    author='Jörg Faschingbauer',

    # https://pypi.org/classifiers/
    classifiers=[
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',
        # 'Intended Audience :: Developers',
        # 'Topic :: Software Development :: Build Tools',
        # 'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
    ],

    keywords='homepage, training, opentraining',  # Optional
    package_dir = {'': 'src'},
    packages = [
        'opentraining',
        'opentraining.sphinxglue',
    ],
    python_requires='>=3.6, <4',
    install_requires=[
        'networkx',
        'Sphinx',
    ],
)



from . import utils
from . import soup
from .. import errors
from ..exercise import Exercise

from sphinx.util.docutils import SphinxDirective
from sphinx.util.nodes import set_source_info
from sphinx.util import logging
from docutils import nodes

logger = logging.getLogger(__name__)


def setup(app):
    app.add_directive('ot-exercise', _ExerciseDirective)
    app.connect('doctree-read', _ev_doctree_read__extract_exercisenodes)

def _ev_doctree_read__extract_exercisenodes(app, doctree):
    try:
        docname = app.env.docname
        exercise_nodes = list(doctree.traverse(_ExerciseNode))
        if len(exercise_nodes) > 1:
            raise errors.OpenTrainingError(f'{docname} contains multiple exercises')

        for n in exercise_nodes:
            soup.sphinx_add_element(app, Exercise(
                title=utils.get_document_title(docname, doctree),
                docname=docname,
                path=n.path, 
                userdata=n,
                dependencies=n.dependencies,
            ))
            n.replace_self([])
    except Exception:
        logger.exception(f'{docname}: cannot extract topic nodes')
        raise
        
class _ExerciseNode(nodes.Element):
    def __init__(self, path, dependencies):
        super().__init__(self)
        self.title = None
        self.path = path
        self.dependencies = dependencies

class _ExerciseDirective(SphinxDirective):
    required_arguments = 1   # path

    option_spec = {
        'dependencies': utils.list_of_elementpath,
    }

    def run(self):
        path = utils.element_path(self.arguments[0].strip())
        dependencies = self.options.get('dependencies', [])

        exercise = _ExerciseNode(path=path, dependencies=dependencies)
        exercise.document = self.state.document
        set_source_info(self, exercise)

        return [exercise]

